/*-----------------------------------------------------------------------------
	tfilesign Copyright (c) TORO
-----------------------------------------------------------------------------*/
#ifndef UNICODE
#define UNICODE
#endif

#include <windows.h>
#include <bcrypt.h>
#include "TOROWIN.H"

#ifndef STATUS_SUCCESS
#define STATUS_SUCCESS ((NTSTATUS)0x00000000L)
#endif

#define UseHashAlgorithm	BCRYPT_SHA1_ALGORITHM
#define UseKeyAlgorithm		BCRYPT_ECDSA_P256_ALGORITHM

BYTE DefaultPublicKey[] = {
 0x45,0x43,0x53,0x31,0x20,0x00,0x00,0x00,
 0x1D,0xA8,0x8F,0xA9,0xB1,0xD7,0x80,0xAD,
 0x4B,0x72,0x1E,0x76,0x60,0xD3,0x8C,0xD7,
 0xB1,0x6A,0x65,0x74,0x72,0x37,0x93,0x06,
 0x43,0x90,0x92,0x8A,0x1D,0x4A,0x76,0x9D,
 0xA8,0x09,0xAB,0x13,0x92,0x54,0xEA,0x36,
 0x3C,0x97,0x1C,0xEB,0x53,0xA9,0x7C,0xD6,
 0xFB,0x72,0x84,0x24,0xFB,0x3A,0x80,0x86,
 0x6D,0x1F,0xFF,0xC1,0xB5,0xD9,0x84,0x3F,
};
#define DefaultPublicKeySize 0x48

const WCHAR PrivateKeyFileName[] = L"tfilesign-privkey.dat";
const WCHAR PublicKeyFileName[]  = L"tfilesign-pubkey.dat";

int MakeKey(const WCHAR *PrivateKeyFileName, const WCHAR *PublicKeyFileName);
int SignFile(const WCHAR *PrivateKeyFileName, const WCHAR *SignSourceFileName, WCHAR *SignedFileName, const WCHAR *command);
BOOL SignData(BCRYPT_KEY_HANDLE hKey, const WCHAR *WriteImagePath, BYTE *Image, DWORD ImageSize, DWORD FixSize);
int VerifyFile(const WCHAR *PublicKeyFileName, const WCHAR *SignedFileName);
BOOL VerifyData(BCRYPT_KEY_HANDLE hKey, BYTE *Image, DWORD ImageSize);

BOOL GetImageHash(BYTE *Image, DWORD ImageSize, BYTE **HashData, DWORD *HashSize);
BOOL ImportKeyData(BCRYPT_ALG_HANDLE hAlg, const WCHAR *KeyFileName, WCHAR *KeyType, BCRYPT_KEY_HANDLE *hKey);

BOOL ExportKeyData(BCRYPT_KEY_HANDLE hKey, const WCHAR *KeyFileName, WCHAR *KeyType);
DWORD LoadFileImage(const WCHAR *filename, DWORD margin, char **image, DWORD *imagesize, DWORD *filesize);

void USEFASTCALL printout(const TCHAR *str);
UTCHAR GetLineParam(LPCTSTR *str, TCHAR *param);



int USECDECL main(void)
{
	TCHAR SourceFileName[MAX_PATH];
	TCHAR SignedFileName[MAX_PATH];
	TCHAR *commandline, exepath[MAX_PATH], command[MAX_PATH];

	commandline = GetCommandLine();
	GetLineParam(&commandline, exepath);
	GetLineParam(&commandline, command);
	GetLineParam(&commandline, SourceFileName);
	GetLineParam(&commandline, SignedFileName);

	switch ( command[0] ){
		case 's': // 
			return SignFile(PrivateKeyFileName, SourceFileName, SignedFileName, command);

		case 'v': // mF
			return VerifyFile(PublicKeyFileName, SourceFileName);

		case 't': // ȊmF
			if ( SourceFileName[0] == '\0' ){
				GetModuleFileName(NULL, SourceFileName, MAX_PATH);
				printout(L"Selftest: ");
			}
			return VerifyFile(NULL, SourceFileName);

		case 'm': // L[쐬
			return MakeKey(PrivateKeyFileName, PublicKeyFileName);

		default:
			printout(
				L"tfilesign Version 1.2 Copyright \xa9" L"2014,2016,2024 TORO\r\n\r\n"
				L"make key    : tfilesign m\r\n"
				L"sign file   : tfilesign s filename [signed filename]\r\n"
				L"verify file : tfilesign v filename\r\n"
				L"test TORO's signature : tfilesign t [filename]\r\n"
			);
	}
	return EXIT_SUCCESS;
}

int MakeKey(const TCHAR *PrivateKeyFileName, const TCHAR *PublicKeyFileName)
{
	BCRYPT_ALG_HANDLE hAlg = NULL;
	BCRYPT_KEY_HANDLE hKey = NULL;
	BOOL success = TRUE;

	if ( STATUS_SUCCESS != BCryptOpenAlgorithmProvider(&hAlg, UseKeyAlgorithm, NULL, 0) ){
		printout( L"Crypt Provider open error.\r\n");
		success = FALSE;
	}

	if ( success ){
		if ( STATUS_SUCCESS != BCryptGenerateKeyPair(hAlg, &hKey, 256, 0) ){
			printout(L"Generate key pair error.\r\n");
			success = FALSE;
		}
	}

	if ( success ){
		BCryptFinalizeKeyPair(hKey, 0);
		ExportKeyData(hKey, PrivateKeyFileName, BCRYPT_ECCPRIVATE_BLOB);
		ExportKeyData(hKey, PublicKeyFileName,  BCRYPT_ECCPUBLIC_BLOB);
	}

	if ( hKey != NULL ) BCryptDestroyKey(hKey);
	if ( hAlg != NULL ) BCryptCloseAlgorithmProvider(hAlg, 0);
	return success ? EXIT_SUCCESS : EXIT_FAILURE;
}

int SignFile(const TCHAR *PrivateKeyFileName, const TCHAR *SignSourceFileName, TCHAR *SignedFileName, const WCHAR *command)
{
	BCRYPT_ALG_HANDLE hAlg = NULL;
	BCRYPT_KEY_HANDLE hKey = NULL;
	BYTE *fileimage = NULL;
	DWORD filesize, fixsize = sizeof(WORD);
	BOOL success = TRUE;

	printout(SignSourceFileName);
	printout(L" : ");
	if ( NO_ERROR != LoadFileImage(SignSourceFileName, 0x1000, &fileimage, NULL, &filesize) ){
		printout(L"file open error.\r\n");
		return EXIT_FAILURE;
	}

	if ( command[1] == 'f' ){
		fixsize = 0;
	}else if ( (*fileimage == 'M') && (*(fileimage + 1) == 'Z') &
		 (*(fileimage + 3) == '\0') ){
		fixsize = 0;

		#define TrimSize 0x84
		if ( (command[1] == 'p') && (filesize > TrimSize) ){
			BYTE *tail;
			int i;

			tail = fileimage + filesize - TrimSize;
			i = TrimSize;
			for ( ;; ){
				if ( *tail++ != '\0' ) break;
				if ( --i > 0  ) continue;
				filesize -= TrimSize;
				break;
			}
		}
	}else if ( (*fileimage != 'P') || (*(fileimage + 1) != 'K') ||
		 (*(fileimage + 2) != '\3') || (*(fileimage + 3) != '\4') ){
		printout(L"file not zip/exe file.\r\n");
		success = FALSE;
	}

	if ( STATUS_SUCCESS != BCryptOpenAlgorithmProvider(&hAlg, UseKeyAlgorithm, NULL, 0) ){

		printout(L"Crypt Provider open error.\r\n");
		success = FALSE;
	}

	if ( success && !ImportKeyData(hAlg, PrivateKeyFileName, BCRYPT_ECCPRIVATE_BLOB, &hKey) ){
		success = FALSE;
	}

	if ( success ){
		WCHAR *p, ext[MAX_PATH];

		if ( *SignedFileName == '\0' ){
			tstrcpy(SignedFileName, SignSourceFileName);
			p = tstrchr(SignedFileName, '.');
			if ( p == NULL ){
				p = SignedFileName + tstrlen(SignedFileName);
				ext[0] = '\0';
			}else{
				tstrcpy(ext, p);
			}
			tstrcpy(p, L"-signed");
			tstrcat(p, ext);
		}

		if ( SignData(hKey, SignedFileName, fileimage, filesize, fixsize) ){
			printout(L"file signed.\r\n");
		}else{
			printout(L"sign error.\r\n");
		}
	}

	if ( hKey != NULL ) BCryptDestroyKey(hKey);
	if ( hAlg != NULL ) BCryptCloseAlgorithmProvider(hAlg, 0);
	HeapFree(GetProcessHeap(), 0, fileimage);
	return success ? EXIT_SUCCESS : EXIT_FAILURE;
}

DWORD DumpSignature(BYTE *dest, BYTE *Signature, DWORD SignatureSize, DWORD FixSize)
{
	DWORD commentsize = SignatureSize * 2 + 3 + 1;

	if ( FixSize == sizeof(WORD) ){
		*(WORD *)dest = (WORD)commentsize;
		dest += sizeof(WORD);
	}
	while ( SignatureSize ){
		*dest = ((*Signature & 0xf0) >> 4) + 'a';
		*(dest + 1) = (*Signature & 0xf) + 'a';
		dest += 2;
		Signature++;
		SignatureSize--;
	}

	*dest = (BYTE)(((commentsize >> 8) & 0xf) + 'a');
	*(dest + 1) = (BYTE)(((commentsize >> 4) & 0xf) + 'a');
	*(dest + 2) = (BYTE)((commentsize & 0xf) + 'a');
	*(dest + 3) = '\0';
	return commentsize + FixSize;
}

BOOL SignData(BCRYPT_KEY_HANDLE hKey, const WCHAR *WriteImagePath, BYTE *Image, DWORD ImageSize, DWORD FixSize)
{
	BYTE *Signature;
	DWORD SignatureSize;
	BYTE *HashData;
	DWORD HashDataSize;
	HANDLE hFile;
	DWORD WriteSize, fsize;
	NTSTATUS status;
	BOOL result;

	ImageSize -= FixSize;
	if ( GetImageHash(Image, ImageSize, &HashData, &HashDataSize) == FALSE ){
		return FALSE;
	}

	BCryptSignHash(hKey, NULL, HashData, HashDataSize, NULL, 0, &SignatureSize, 0);
	Signature = (LPBYTE)HeapAlloc(GetProcessHeap(), 0, SignatureSize);
	status = BCryptSignHash(hKey, NULL, HashData, HashDataSize, Signature, SignatureSize, &SignatureSize, 0);
	if ( STATUS_SUCCESS != status ){
		HeapFree(GetProcessHeap(), 0, HashData);
		HeapFree(GetProcessHeap(), 0, Signature);
		return FALSE;
	}
	fsize = ImageSize + DumpSignature(Image + ImageSize, Signature, SignatureSize, FixSize);

	hFile = CreateFile(WriteImagePath, GENERIC_WRITE, 0, NULL, CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL);
	if ( hFile == INVALID_HANDLE_VALUE ){
		printout(L"file open error.\r\n");
		result = FALSE;
	}
	result = WriteFile(hFile, Image, fsize, &WriteSize, NULL);
	CloseHandle(hFile);

	HeapFree(GetProcessHeap(), 0, HashData);
	HeapFree(GetProcessHeap(), 0, Signature);
	return result;
}

int VerifyFile(const TCHAR *PublicKeyFileName, const TCHAR *SignedFileName)
{
	BCRYPT_ALG_HANDLE hAlg = NULL;
	BCRYPT_KEY_HANDLE hKey = NULL;
	BYTE *fileimage = NULL;
	DWORD filesize;
	BOOL success = TRUE;

	printout(SignedFileName);
	printout(L" : ");
	if ( NO_ERROR != LoadFileImage(SignedFileName, 0x1000, &fileimage, NULL, &filesize) ){
		printout(L"file open error.\r\n");
		return EXIT_FAILURE;
	}

	if ( STATUS_SUCCESS != BCryptOpenAlgorithmProvider(&hAlg, UseKeyAlgorithm, NULL, 0) ){
		printout(L"Crypt Provider open error.\r\n");
		success = FALSE;
	}

	if ( success && !ImportKeyData(hAlg, PublicKeyFileName, BCRYPT_ECCPUBLIC_BLOB, &hKey) ){
		printout(L"key.dat load error.\r\n");
		success = FALSE;
	}

	if ( success ){
		if ( VerifyData(hKey, fileimage, filesize) ){
			printout(L"file signature is OK.\r\n");
		}else{
			printout(L"file signature error.\r\n");
			success = FALSE;
		}
	}

	if ( hKey != NULL ) BCryptDestroyKey(hKey);
	if ( hAlg != NULL ) BCryptCloseAlgorithmProvider(hAlg, 0);
	HeapFree(GetProcessHeap(), 0, fileimage);
	return success ? EXIT_SUCCESS : EXIT_FAILURE;
}

BOOL VerifyData(BCRYPT_KEY_HANDLE hKey, BYTE *Image, DWORD ImageSize)
{
	DWORD DataSize;
	BYTE *Signature;
	DWORD SignatureSize;
	BYTE *HashData;
	DWORD HashDataSize;
	BYTE *src, *dest;
	DWORD left;
	NTSTATUS status;

	if ( ImageSize < (DWORD)0x90 ) return FALSE;
	SignatureSize =	((*(Image + ImageSize - 4) - 'a') << 8) +
					((*(Image + ImageSize - 3) - 'a') << 4) +
					 (*(Image + ImageSize - 2) - 'a');
	if ( (SignatureSize < 0x10) || (SignatureSize > ImageSize) || (SignatureSize >= 0x100) ){
		return FALSE;
	}

	DataSize = ImageSize - SignatureSize;
	if ( *Image == 'P' ) DataSize -= sizeof(WORD);
	Signature = Image + ImageSize - SignatureSize;
	SignatureSize = (SignatureSize - 3 - 1) / 2;

	src = dest = Signature;
	left = SignatureSize;
	while ( left ){
		*dest++ = ((*src - 'a') << 4) + (*(src + 1) - 'a');
		src += 2;
		left--;
	}

	if ( GetImageHash(Image, DataSize, &HashData, &HashDataSize) == FALSE ){
		return FALSE;
	}

	status = BCryptVerifySignature(hKey, NULL, HashData, HashDataSize, Signature, SignatureSize, 0);
	HeapFree(GetProcessHeap(), 0, HashData);

	return (status == STATUS_SUCCESS);
}

BOOL GetImageHash(BYTE *Image, DWORD ImageSize, BYTE **HashData, DWORD *HashSize)
{
	BCRYPT_ALG_HANDLE hAlg;
	BCRYPT_HASH_HANDLE hHash;
	DWORD resultsize;
	DWORD worksize;
	BYTE *hashwork;

	if ( STATUS_SUCCESS != BCryptOpenAlgorithmProvider(&hAlg, UseHashAlgorithm, NULL, 0) ){
		return FALSE;
	}

	BCryptGetProperty(hAlg, BCRYPT_OBJECT_LENGTH, (LPBYTE)&worksize, sizeof(worksize), &resultsize, 0);
	hashwork = (BYTE *)HeapAlloc(GetProcessHeap(), 0, worksize);
	if ( STATUS_SUCCESS != BCryptCreateHash(hAlg, &hHash, hashwork, worksize, NULL, 0, 0) ){
		HeapFree(GetProcessHeap(), 0, hashwork);
		BCryptCloseAlgorithmProvider(hAlg, 0);
		return FALSE;
	}
	BCryptHashData(hHash, Image, ImageSize, 0);
	BCryptGetProperty(hAlg, BCRYPT_HASH_LENGTH, (LPBYTE)HashSize, sizeof(DWORD), &resultsize, 0);
	*HashData = (LPBYTE)HeapAlloc(GetProcessHeap(), 0, *HashSize);
	BCryptFinishHash(hHash, *HashData, *HashSize, 0);

	BCryptDestroyHash(hHash);
	HeapFree(GetProcessHeap(), 0, hashwork);
	BCryptCloseAlgorithmProvider(hAlg, 0);
	return TRUE;
}

void GetFilePath(TCHAR *path, const TCHAR *filename)
{
	tstrcpy(path, filename);
	if ( tstrchr(path, '\\') == NULL ){
		GetModuleFileName(NULL, path, MAX_PATH);
		tstrcpy(tstrrchr(path, '\\') + 1, filename);
	}
}

BOOL ImportKeyData(BCRYPT_ALG_HANDLE hAlg, const WCHAR *KeyFileName, WCHAR *KeyType, BCRYPT_KEY_HANDLE *hKey)
{
	BYTE *keydata = NULL;
	DWORD keysize;
	TCHAR filepath[MAX_PATH];
	NTSTATUS status;

	if ( KeyFileName == NULL ){
		return (BCryptImportKeyPair(hAlg, NULL, KeyType, hKey, DefaultPublicKey, DefaultPublicKeySize, 0) == STATUS_SUCCESS);
	}

	GetFilePath(filepath, KeyFileName);
	if ( NO_ERROR != LoadFileImage(filepath, 0, &keydata, NULL, &keysize) ){
		printout(L"keyfile open error.\r\n");
		return FALSE;
	}
	status = BCryptImportKeyPair(hAlg, NULL, KeyType, hKey, keydata, keysize, 0);
	HeapFree(GetProcessHeap(), 0, keydata);
	return (status == STATUS_SUCCESS);
}

BOOL ExportKeyData(BCRYPT_KEY_HANDLE hKey, const WCHAR *KeyFileName, WCHAR *KeyType)
{
	BYTE *keydata;
	DWORD keysize;
	TCHAR filepath[MAX_PATH];
	HANDLE hFile;
	DWORD writesize;
	BOOL result = FALSE;

	BCryptExportKey(hKey, NULL, KeyType, NULL, 0, &keysize, 0);
	keydata = (BYTE *)HeapAlloc(GetProcessHeap(), 0, keysize);
	if ( STATUS_SUCCESS == BCryptExportKey(hKey, NULL, KeyType, keydata, keysize, &writesize, 0) ){
		GetFilePath(filepath, KeyFileName);
		hFile = CreateFile(filepath, GENERIC_WRITE, 0, NULL, CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL);
		if ( hFile != INVALID_HANDLE_VALUE ){
			result = WriteFile(hFile, keydata, keysize, &writesize, NULL);
			CloseHandle(hFile);
		}else{
			printout(L"keyfile open error.\r\n");
		}
	}

	if ( KeyType == BCRYPT_ECCPUBLIC_BLOB ){
		BYTE *dumpptr = keydata;
		DWORD dumpsize = 0;

		printout(L"BYTE DefaultPublicKey[] = {");

		for ( ; dumpsize < keysize ; dumpsize++ ){
			if ( (dumpsize & 7) == 0 ) printout(L"\r\n ");
			wsprintf(filepath, L"0x%02X,", *dumpptr++);
			printout(filepath);
		}
		wsprintf(filepath, L"\r\n}\r\n#define DefaultPublicKeySize 0x%X\r\n", keysize);
		printout(filepath);
	}

	HeapFree(GetProcessHeap(), 0, keydata);
	return result;
}

DWORD LoadFileImage(const WCHAR *filename, DWORD margin, char **image, DWORD *imagesize, DWORD *filesize)
{
	BOOL useheap; // q[vmۂȂ true
	HANDLE hFile;
	DWORD imgsize;
	DWORD sizeL, sizeH;		// t@C̑傫
	DWORD result;
	HWND hWnd = NULL;
										// t@CJ ---------------------
	hFile = CreateFileW(filename, GENERIC_READ,
			FILE_SHARE_WRITE | FILE_SHARE_READ, NULL,
			OPEN_EXISTING, FILE_FLAG_SEQUENTIAL_SCAN, NULL);
	if ( hFile == INVALID_HANDLE_VALUE ) return GetLastError();
										// t@CTCY̊mF ---------------
	sizeL = GetFileSize(hFile, &sizeH);
	if ( sizeH ) sizeL = FFD;
	if ( filesize ) *filesize = sizeL;
										// ǂݍݏ -----------------------
	if ( *image ){				// ͊mۍς
		useheap = FALSE;
		imgsize = *imagesize;
		if ( imgsize < sizeL ) sizeL = imgsize;
		if ( imgsize < (sizeL + margin) ) margin = imgsize - sizeL;
	}else{						// m
		useheap = TRUE;
		if ( (*image = HeapAlloc(GetProcessHeap(), 0, sizeL + margin)) == NULL ){
			result = GetLastError();
			CloseHandle(hFile);
			return result;
		}
	}
										// ǂݍ & 0 padding ---------------
	result = ReadFile(hFile, *image, sizeL, &sizeL, NULL) ?
			NO_ERROR : GetLastError();
	CloseHandle(hFile);
	if ( result != NO_ERROR ){
		if ( useheap != FALSE ) HeapFree(GetProcessHeap(), 0, *image);
		return result;
	}
	if ( margin != 0 ) memset(*image + sizeL, 0, margin);
	if ( imagesize != NULL ) *imagesize = sizeL;
	return NO_ERROR;
}

void GetQuotedParameter(LPCTSTR *commandline, TCHAR *param, const TCHAR *parammax)
{
	const TCHAR *ptr, *ptrfirst, *ptrlast;
	TCHAR *dest;

	dest = param;
	ptrfirst = ptr = *commandline + 1;
	for ( ; ; ){
		TCHAR code;

		code = *ptr;
		if ( (code == '\0') || (code == '\r') || (code == '\n') ){
			ptrlast = ptr;
			break;
		}
		if ( code != '\"' ){
			ptr++;
			continue;
		}
		// " ꍇ̏
		if ( *(ptr + 1) != '\"' ){	// "" GXP[v?
			ptrlast = ptr++; // P " c ŏI
			break;
		}
		// GXP[v
		{
			size_t copysize;

			copysize = ptr - ptrfirst + 1;
			if ( (dest + copysize) >= parammax ){ // buffer overflow?
				// "  1GXP[v̂ ">=" ok
				ptrlast = ptr;
				break;
			}
			memcpy(dest, ptrfirst, TSTROFF(copysize));
			dest += copysize;
			ptrfirst = (ptr += 2); // " x 2
			continue;
		}
	}
	*commandline = ptr;
	{
		size_t ptrsize;

		ptrsize = ptrlast - ptrfirst;
		if ( (dest + ptrsize) > parammax ){ // buffer overflow?
			tstrcpy(param, T("<flow!!>"));
		}else{
			memcpy(dest, ptrfirst, TSTROFF(ptrsize));
			*(dest + ptrsize) = '\0';
		}
	}
}

/*-------------------------------------
	s(\0, \a, \d)H

RET:	==0 :EOL	!=0 :Code
-------------------------------------*/
UTCHAR USEFASTCALL IsEOL(LPCTSTR *str)
{
	UTCHAR code;

	code = **str;
	if ( code == 0 ) return 0;
	if ( code == 0xd ){
		if ( *(*str + 1) == 0x0a ) (*str)++;
		return 0;
	}
	if ( code == 0xa ){
		if ( *(*str + 1) == 0x0d ) (*str)++;
		return 0;
	}
	return code;
}

/*-----------------------------------------------------------------------------
	̃p[^𒊏oB擪Ɩ̋󔒂͏

RET:	擪̕(Ȃ 0)
-----------------------------------------------------------------------------*/
UTCHAR GetLineParam(LPCTSTR *str, TCHAR *param)
{
	UTCHAR code, bottom;
	const TCHAR *src;
	TCHAR *dst, *max;

	src = *str;
	dst = param;
	max = dst + MAX_PATH - 1;
	for ( ;; ){
		code = *src;
		if ( (code == ' ') || (code == '\t') || (code == '\r') || (code == '\n') ){
			src++;
			continue;
		}
		break;
	}
	bottom = code;
	if ( code == '\0' ){
		*str = src;
		*dst = '\0';
		return '\0';
	}
	if ( code == '\"' ){
		GetQuotedParameter(&src, dst, max);
	}else{
		do {
			src++;
			if ( code <= ' ' )	break;
			if ( code == '\"' ){
				*dst++ = code;
				while( '\0' != (code = IsEOL(&src)) ){
					src++;
					*dst++ = code;
					if ( (code == '\"') && (*src != '\"') ) break;
				}
			}else{
				*dst++ = code;
			}
			if ( dst >= max ) break;
		}while( '\0' != (code = IsEOL(&src)) );
		*dst = '\0';
	}
	*str = src;
	return bottom;
}

/*-----------------------------------------------------------------------------
	R\[ps\
-----------------------------------------------------------------------------*/
void USEFASTCALL printout(const TCHAR *str)
{
	DWORD size;

	WriteConsole(GetStdHandle(STD_OUTPUT_HANDLE), str, (DWORD)tstrlen(str), &size, NULL);
}
